"""
client.py

makes calls to terraform CLI
"""

import subprocess
from pathlib import Path
from typing import List


class TfClient:
    """
    CLI client for terraform
    """

    def __init__(self, work_dir: Path):
        self.work_dir = work_dir

    def state_pull(self, state_file: Path):
        """Execute state pull and redirect to the given state file"""
        self._run_command("state pull", f'> "{state_file}"')

    def state_push(self, state_file: Path):
        """Execute state push for the given state file"""
        self._run_command("state push", f'"{state_file}"')

    def state_list(self, state_file: Path) -> List[str]:
        """List the content of the state (from the state file if given and exists)"""
        opt = f"-state={state_file}" if state_file.exists() else ""
        return self._run_command("state list", opt).split()

    # pylint: disable=too-many-arguments
    def state_mv(self, from_file: Path, to_file: Path, from_object: str, to_object: str, opts: str = ""):
        """Execute state mv"""
        self._run_command("state mv", f'-state={from_file} -state-out="{to_file}" {from_object} {to_object} {opts}')

    def _run_command(self, command: str, extra="") -> str:
        """Execute a terraform command in the work directory"""
        cmd = f"""
        cd "{self.work_dir}" ;
        terraform state {command} {extra};
        """
        result = subprocess.run(["/usr/bin/bash", "-c", cmd], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                check=True)
        if result.returncode != 0:
            raise RuntimeError(f"'terraform {cmd}' command failed: {result.stderr.decode('unicode_escape')}")
        return result.stdout.decode('unicode_escape')



"""
state.py

related to terraform state operations
"""

from pathlib import Path
from typing import Union, Optional, List, Tuple

from tfmv.client import TfClient
from tfmv.error import DuplicateError


class TfState:
    """
    represents a terraform state
    """

    def __init__(self, path: Union[str, Path]):
        self.home_path = Path(path)
        self._local_state_path: Optional[Path] = None
        self._tf_client = TfClient(self.home_path)

    @property
    def local_state_path(self) -> Path:
        """Path of the local state file"""
        if self._local_state_path is None:
            return self.home_path.joinpath("terraform.tfstate")
        return self._local_state_path

    def pull(self, state_file: Path = None) -> None:
        """Pulls the remote state to a local file (created in home directory if none is given)"""
        if state_file is not None:
            self._local_state_path = state_file
        if self.local_state_path.exists():
            raise FileExistsError(f"state '{self.home_path}' already has a local file, call `clean` first")
        self._tf_client.state_pull(self.local_state_path)

    def push(self) -> None:
        """Pushes the local state file to the remote state"""
        if not self.local_state_path.exists():
            raise FileNotFoundError(f"state '{self.home_path}' has no local file, call `pull` first")
        self._tf_client.state_push(self.local_state_path)

    def clean(self, dry_run=False) -> None:
        """Clean the local state file and its backups"""
        if not self.local_state_path.exists():
            return
        state_dir = self.local_state_path.parent
        for file in state_dir.glob(f"{self.local_state_path.name}*"):
            if dry_run:
                print(f"would delete file: {file}")
            else:
                file.unlink()

    def list(self) -> List[str]:
        """list the items in the state"""
        return self._tf_client.state_list(self.local_state_path)

    def move_to(self, from_object: str, to_object: str, dest: 'TfState', dry_run=False):
        """move an item"""
        if not self.local_state_path.exists():
            raise FileNotFoundError(f"source state '{self.home_path}' has no local state file, call `pull` first")
        if not dest.local_state_path.exists():
            raise FileNotFoundError(f"dest state '{self.home_path}' has no local state file, call `pull` first")
        opts = "-dry-run" if dry_run else ""
        self._tf_client.state_mv(self.local_state_path, dest.local_state_path, from_object, to_object, opts)


def filter_and_rename(items: List[str], prefix: str, new_prefix: str) -> Tuple[List[str], List[str]]:
    """filter items by prefix, and return replaces the prefix by a new prefix
    :param items: initial list of items
    :param prefix: only items starting with prefix + '.', or equal to prefix, are selected
    :param new_prefix: new prefix for renaming
    :return: tuple (filtered_items, renamed_items)
    """
    # """"""
    # prefix = f"{prefix}."
    from_items, to_items, duplicates = [], [], []
    for item in items:
        if item == prefix or item.startswith(f'{prefix}.'):
            new_item = f'{new_prefix}{item[len(prefix):]}'
            if new_item.startswith('.'):
                new_item = new_item[1:]
            from_items.append(item)
            if new_item in items and new_prefix != prefix:
                duplicates.append(new_item)
            else:
                to_items.append(new_item)
    if any(duplicates):
        raise DuplicateError(*duplicates)
    return from_items, to_items



"""
TFMV

simple terraform utility
"""
import tempfile
from pathlib import Path

import click

from tfmv.state import TfState, filter_and_rename

VALID_PATH = click.Path(exists=True, file_okay=False, writable=True)


@click.command()
@click.argument("from_", metavar="FROM")
@click.argument("to_", metavar="TO", default="", required=False)
@click.option("--src", help="source terraform configuration directory", default=".", type=VALID_PATH)
@click.option("--dest", help="destination terraform configuration directory (same as src if omitted)", default=None,
              type=VALID_PATH)
@click.option("--dry-run", "-n", help="do no write operation", is_flag=True, default=False)
@click.option("--verbose", "-v", is_flag=True, help="show detailed output")
@click.option("--use-tmp-dir", is_flag=True,
              help="use a temporary directory for local state files, instead of the configuration directory")
# pylint: disable=invalid-name
def mv(*args, **kwargs):
    """Terraform tool for copying/moving objects in/between states.

    Objects in the source state starting with the prefix `FROM` wil be used.
    If a value for `TO` is specified, the prefix will be changed to this value
    """
    _mv(*args, **kwargs)


# pylint: disable=too-many-arguments, invalid-name, too-many-locals, too-many-branches
def _mv(from_, to_="", src=".", dest=None, dry_run=False, verbose=False, use_tmp_dir=False):
    """implementation of the mv command"""
    if verbose:
        click.echo(f"from    = {from_}")
        click.echo(f"to      = {to_}")
        click.echo(f"src     = {src}")
        click.echo(f"dest    = {dest}")

    # check and adapt parameters
    if dest is None and (to_ in ("", from_)):
        raise click.BadParameter("cannot move in place! (set TO or --dest or both)")
    src_state = TfState(src)
    dest_state = TfState(dest) if dest is not None else src_state
    if to_ == "":
        to_ = from_

    tmp = None
    src_state_file = None
    dest_state_file = None

    try:
        # get temporary local files if required
        if use_tmp_dir:
            tmp = tempfile.TemporaryDirectory()
            tmp_dir = Path(tmp.name)
            src_state_file = tmp_dir.joinpath("src-terraform.tfstate")
            dest_state_file = tmp_dir.joinpath("dest-terraform.tfstate")

        # pull states and read objects
        src_state.pull(src_state_file)
        if dest is not None:
            dest_state.pull(dest_state_file)
        src_objects = src_state.list()

        # filter objects and get new names
        from_objects, to_objects = filter_and_rename(src_objects, from_, to_)

        # log move info
        header = "would move: " if dry_run else "moving: "
        no_header = " " * len(header)
        if verbose:
            _print_op("", "src workspace:  ", src_state.home_path, color='white')
            _print_op("", "dest workspace: ", dest_state.home_path, color='white')
            _print_op("", "src state:  ", src_state.local_state_path, color='white')
            _print_op("", "dest state: ", dest_state.local_state_path, color='white')

        # move selected objects
        for from_object, to_object in zip(from_objects, to_objects):
            _print_op(header, "from: ", from_object)
            _print_op(no_header, "to:   ", to_object)
            src_state.move_to(from_object=from_object, to_object=to_object, dest=dest_state, dry_run=dry_run)

        # push state to remote
        if dry_run:
            click.echo(click.style("would push states to remote"))
        else:
            click.echo(click.style("pushing modified states to remote"))
            src_state.push()
            if dest:
                dest_state.push()

        # clean local state files
        src_state.clean(dry_run)
        if dest is not None:
            dest_state.clean(dry_run)

    finally:
        # clean temporary files
        if tmp is not None:
            tmp.cleanup()


def _print_op(row_header, relation, element, color="cyan"):
    click.echo(click.style(row_header, fg='yellow'), nl=False)
    click.echo(click.style(relation, fg=f'bright_{color}'), nl=False)
    click.echo(click.style(f"{element}", fg=f'{color}'))



"""
error.py

classes for exceptions
"""


class DuplicateError(Exception):
    """exception raised when state operations would create duplicates"""
    def __init__(self, *duplicates: str):
        super().__init__()
        self.duplicates = list(duplicates)



import os
import sys

import pipfile
from setuptools import setup, find_packages

##########
# version
##########

try:
    with open(os.path.join(os.path.dirname(__file__), 'version.txt'), 'r') as file:
        VERSION = "".join(file.readlines()).strip()
except IOError:
    VERSION = None
VERSION = os.environ.get('VERSION', VERSION)  # use environment variable if present (overrides file)
if VERSION is None:
    print("version is not specified! use 'version.txt' file at project root, or environment variable VERSION")
    sys.exit(1)


##########
# setup
##########

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

INSTALL_REQUIRES = pipfile.load('Pipfile').data['default'].keys()

setup(
    name="tfmv",
    version=VERSION,
    author='Vincent Pfister',
    author_email='vincent.pfister@raisepartner.com',
    description='Simple tool for moving / renaming resources in one or several terraform configurations.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url="https://gitlab.raisepartner.com/it_group/infra/terraform/tfmv",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    entry_points="""
        [console_scripts]
        tfmv=tfmv.main:mv
    """
)
